package cn.com.qmth.examcloud.core.oe.student.api.controller;

import cn.com.qmth.examcloud.api.commons.security.bean.User;
import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.core.oe.student.base.OeConstants;
import cn.com.qmth.examcloud.core.oe.student.base.utils.Check;
import cn.com.qmth.examcloud.core.oe.student.bean.*;
import cn.com.qmth.examcloud.core.oe.student.bean.client.CourseInfo;
import cn.com.qmth.examcloud.core.oe.student.service.*;
import cn.com.qmth.examcloud.starters.crypto.common.CryptoConstant;
import cn.com.qmth.examcloud.support.CacheConstants;
import cn.com.qmth.examcloud.support.Constants;
import cn.com.qmth.examcloud.support.enums.FileAnswerAcknowledgeStatus;
import cn.com.qmth.examcloud.support.examing.ExamFileAnswer;
import cn.com.qmth.examcloud.support.examing.ExamRecordData;
import cn.com.qmth.examcloud.support.fss.FssFactory;
import cn.com.qmth.examcloud.support.fss.FssHelper;
import cn.com.qmth.examcloud.support.fss.model.FssSignInfo;
import cn.com.qmth.examcloud.support.redis.RedisKeyHelper;
import cn.com.qmth.examcloud.web.helpers.SequenceLockHelper;
import cn.com.qmth.examcloud.web.redis.RedisClient;
import cn.com.qmth.examcloud.web.support.ControllerSupport;
import cn.com.qmth.examcloud.web.support.Naked;
import com.mysql.cj.util.StringUtils;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import io.swagger.annotations.ApiParam;
import org.apache.commons.lang.math.RandomUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.web.bind.annotation.*;

import javax.servlet.http.HttpServletRequest;
import javax.validation.Valid;

@Api(tags = "在线考试控制")
@RestController
@RequestMapping("${app.api.oe.student}/examControl")
public class ExamControlController extends ControllerSupport {

    private static final Logger log = LoggerFactory.getLogger(ExamControlController.class);

    private static final String SEPARATOR = "/";

    private static final String UNDERLINE = "_";

    // 小程序答案上传目录
    private static final String OE_ANSWER_FILE_PATH = "oe-answer-file";

    @Autowired
    private ExamControlService examControlService;

    @Autowired
    private ExamFileAnswerService examFileAnswerService;

    @Autowired
    private RedisClient redisClient;

    @Autowired
    private ExamRecordDataService examRecordDataService;

    @Autowired
    private ExamRecordQuestionsService examRecordQuestionsService;

    @Autowired
    private TempConfigService tempConfigService;

    /**
     * 开始考试
     */
    @Deprecated
    @ApiOperation(value = "开始考试（待废弃）")
    @GetMapping("/startExam")
    public StartExamInfo startExam(@RequestParam Long examStudentId, HttpServletRequest request) {
        User user = getAccessUser();
        if (tempConfigService.checkDisabledForOldApi(user.getRootOrgId())) {
            // 判断学校是否被禁用旧版本API
            throw new StatusException("403X01", "禁止访问！");
        }

        String fromBy = "startExam-oldApi";
        return examControlService.startExam(examStudentId, user.getUserId(), getIp(request), null, fromBy);
    }

    /**
     * 开始考试（新）
     */
    @ApiOperation(value = "开始考试（不支持“ONLINE”考试类型）")
    @PostMapping("/unonline/startExam")
    public StartExamInfo startExamForOther(@RequestParam Long examStudentId, HttpServletRequest request) {
        User user = getAccessUser();
        String fromBy = "startExam-unonline-webClient";
        return examControlService.startExam(examStudentId, user.getUserId(), getIp(request), false, fromBy);
    }

    /**
     * 开始考试（新）
     * 注：需要加、解密
     */
    @ApiOperation(value = "开始考试（仅支持“ONLINE”考试类型）")
    @PostMapping(value = "/online/startExam", consumes = MediaType.TEXT_PLAIN_VALUE, produces = MediaType.TEXT_PLAIN_VALUE)
    public String startExamWithCrypto(@RequestBody String encryptParams, HttpServletRequest request) {
        User user = getAccessUser();
        String fromBy = "startExam-online-webClient";
        return examControlService.startExamWithCrypto(encryptParams, user, getIp(request),
                request.getHeader(CryptoConstant.TIMESTAMP), fromBy);
    }

    /**
     * 开始答题
     */
    @ApiOperation(value = "开始答题")
    @PostMapping("/startAnswer")
    public StartAnswerInfo startAnswer(@RequestParam @ApiParam(value = "考试记录id") Long examRecordDataId) {
        User user = getAccessUser();
        return examControlService.startAnswer(examRecordDataId, user.getUserId());
    }

    /**
     * 考生作答（新）
     * 注：需要加、解密
     */
    @ApiOperation(value = "考试过程中-考生试题作答（新）")
    @PostMapping("/submitQuestionAnswer")
    public void submitQuestionAnswerWithCrypto(@RequestBody ExamStudentQuestionAnswerInfo data, HttpServletRequest request) {
        User user = getAccessUser();
        String referer = request.getHeader("REFERER");
        String agent = request.getHeader("USER-AGENT");
        String timestampStr = request.getHeader(CryptoConstant.TIMESTAMP);
        String fromBy = "submitQuestionAnswer-webClient";
        examRecordQuestionsService.submitQuestionAnswerWithCrypto(data, user, referer, agent, timestampStr, getIp(request), fromBy);
    }

    /**
     * 断点续考:检查正在进行中的考试
     */
    @ApiOperation(value = "断点续考:检查正在进行中的考试")
    @GetMapping("/checkExamInProgress")
    public ExamProcessResultInfo checkExamInProgress(HttpServletRequest request) {
        User user = getAccessUser();
        String fromBy = "checkExamInProgress-webClient";
        return examControlService.checkExamInProgress2(user.getUserId(), getIp(request), fromBy);
    }

    /**
     * 考试心跳
     *
     * @return 剩余时间
     */
    @ApiOperation(value = "考试心跳")
    @GetMapping("/examHeartbeat")
    public Long examHeartbeat(HttpServletRequest request) {
        User user = getAccessUser();
        String fromBy = "examHeartbeat-webClient";
        return examControlService.examHeartbeat(user, getIp(request), fromBy);
    }

    /**
     * 结束考试：交卷
     */
    @Deprecated
    @ApiOperation(value = "结束考试：交卷（待废弃）")
    @GetMapping("/endExam")
    public void endExam(@RequestParam(required = false) Boolean force, HttpServletRequest request) {
        User user = getAccessUser();
        String fromBy = "handInExam-manual-oldApi";
        examControlService.manualEndExam(user.getUserId(), getIp(request), force, fromBy);
    }

    /**
     * 结束考试：交卷（新）
     * 注：需要加、解密
     */
    @ApiOperation(value = "结束考试：交卷（新）")
    @PostMapping(value = "/endExam", consumes = MediaType.TEXT_PLAIN_VALUE)
    public void endExamWithCrypto(@RequestBody String encryptParams,
                                  @RequestParam(required = false) Boolean force,
                                  HttpServletRequest request) {
        User user = getAccessUser();
        String fromBy = "handInExam-manual-webClient";
        examControlService.manualEndExamWithCrypto(encryptParams, user, getIp(request), force,
                request.getHeader(CryptoConstant.TIMESTAMP), fromBy);
    }

    @ApiOperation(value = "校验交卷冻结时间")
    @PostMapping(value = "check/time")
    public void checkTime() {
        User user = getAccessUser();
        examControlService.checkAndComputeExamDuration(user.getUserId(), false);
    }

    /**
     * 获取考试记录信息
     */
    @ApiOperation(value = "获取考试记录信息")
    @GetMapping("/getEndExamInfo")
    public EndExamInfo getEndExamInfo(@RequestParam Long examRecordDataId) {
        return examControlService.getEndExamInfo(examRecordDataId);
    }

    @Deprecated
    @ApiOperation(value = "获取作答文件上传签名-又拍云（微信小程序调用，废弃）")
    @PostMapping("/upyunSignature")
    public FssSignInfo getUpyunSignature(@ModelAttribute @Valid GetYunSignatureReq req) {
        return getAnswerUploadSign(req);
    }

    @ApiOperation(value = "获取作答文件上传签名（微信小程序调用）")
    @PostMapping("/yunSignature")
    public FssSignInfo getYunSignature(@ModelAttribute @Valid GetYunSignatureReq req) {
        return getAnswerUploadSign(req);
    }

    private FssSignInfo getAnswerUploadSign(GetYunSignatureReq req) {
        String fileSuffix = req.getFileSuffix();
        if (StringUtils.isNullOrEmpty(fileSuffix)) {
            throw new StatusException("5002", "文件后缀名不允许为空");
        }
        fileSuffix = !fileSuffix.contains(".") ? "." + fileSuffix : fileSuffix;

        Long examStudentId;
        // 是否为环境监测场景
        boolean isTestDev = Constants.OE_ENV_CHECK_EXAM_RECORD_DATA_ID.equals(req.getExamRecordDataId());
        if (isTestDev) {
            examStudentId = Constants.OE_ENV_CHECK_EXAM_STUDENT_ID;
        } else {
            ExamRecordData examRecordData = examRecordDataService.getExamRecordDataCache(req.getExamRecordDataId());
            if (examRecordData == null) {
                log.warn("ExamRecordDataCache not exist! examRecordDataId = {}", req.getExamRecordDataId());
                throw new StatusException("5002", "请求参数值有误！");
            }
            examStudentId = examRecordData.getExamStudentId();
        }

        // 路径：/oe-answer-file/{examStudentId}/{examRecordDataId}/{order}/{examStudentId}_{examRecordDataId}_{order}_{timeMillis}{randomNum}.jpg",
        StringBuilder filePath = new StringBuilder();
        filePath.append(OE_ANSWER_FILE_PATH).append(SEPARATOR)
                .append(examStudentId).append(SEPARATOR).append(req.getExamRecordDataId())
                .append(SEPARATOR).append(req.getOrder()).append(SEPARATOR)
                .append(examStudentId).append(UNDERLINE).append(req.getExamRecordDataId())
                .append(UNDERLINE).append(req.getOrder()).append(UNDERLINE).append(System.currentTimeMillis())
                .append(RandomUtils.nextInt(8999) + 1000);

        if (!StringUtils.isNullOrEmpty(req.getExt())) {
            filePath.append(UNDERLINE).append(req.getExt());
        }
        filePath.append(fileSuffix);

        FssSignInfo result = FssFactory.getInstance(false).buildSign(filePath.toString());

        String signIdentifier = String.valueOf(System.currentTimeMillis());
        result.setSignIdentifier(signIdentifier);

        return result;
    }

    /**
     * 校验二维码（微信小程序调用）
     */
    @Naked
    @ApiOperation(value = "校验二维码（微信小程序调用）")
    @PostMapping("/checkQrCode")
    public CheckQrCodeInfo checkQrCode(@RequestParam(required = true) String qrCode) {
        return examControlService.checkQrCode(qrCode);
    }

    /**
     * 保存上传的文件（微信小程序调用）
     */
    @ApiOperation(value = "保存上传的文件（微信小程序调用）")
    @PostMapping("/saveUploadedFile")
    public String saveUploadedFile(@ModelAttribute @Valid SaveUploadedFileReq req) {
        User user = getAccessUser();
        ExamFileAnswer examFileAnswer = new ExamFileAnswer();
        examFileAnswer.setExamRecordDataId(req.getExamRecordDataId());
        examFileAnswer.setExamStudentId(req.getExamStudentId());
        examFileAnswer.setQuestionOrder(req.getOrder());
        examFileAnswer.setFilePath(req.getFilePath());
        examFileAnswer.setStatus(FileAnswerAcknowledgeStatus.UNCONFIRMED);
        examFileAnswer.setTransferFileType(req.getTransferFileType());
        String fileAnswerId = RedisKeyHelper.getBuilder().studentFileAnswerKey(req.getExamRecordDataId(), req.getOrder());
        examFileAnswerService.saveFileAnswer(fileAnswerId, examFileAnswer);

        try {
            String fileUrl = "";
            if (req.getFilePath().indexOf(",") > -1) {
                for (String url : req.getFilePath().split(",")) {
                    fileUrl += FssHelper.finalFileUrl(url) + ",";
                }
                fileUrl = fileUrl.substring(0, fileUrl.length() - 1);
            } else {
                fileUrl = FssHelper.finalFileUrl(req.getFilePath());
            }
            examControlService.sendFileAnswerToWebSocket(req.getExamRecordDataId(), req.getOrder(),
                    fileUrl, req.getTransferFileType(), user.getUserId(), user.getRootOrgId());
        } catch (Exception e) {
            examFileAnswerService.deleteFileAnswer(fileAnswerId);
            throw new StatusException("100009", "消息通知失败", e);
        }
        return fileAnswerId;
    }


    /**
     * 查询客户端对上传的文件的响应状态（微信小程序调用）
     */
    @ApiOperation(value = "查询客户端对上传的文件的响应状态（微信小程序调用）")
    @PostMapping("/getUploadedFileAcknowledgeStatus")
    public String getUploadedFileAcknowledgeStatus(@ModelAttribute @Valid GetUploadedFileAcknowledgeStatusReq req) {
        ExamFileAnswer fileAnswer = examFileAnswerService.getFileAnswer(req.getAcknowledgeId());
        if (null != fileAnswer) {
            return fileAnswer.getStatus().toString();
        }
        return FileAnswerAcknowledgeStatus.UNCONFIRMED.toString();
    }

    /**
     * 修改上传音频结果推送状态
     */
    @ApiOperation(value = "修改上传音频结果推送状态")
    @PostMapping("/saveUploadedFileAcknowledgeStatus")
    public void saveUploadedFileAcknowledgeStatus(@RequestBody @Valid SaveUploadedFileAcknowledgeStatusReq req) {
        String acknowledgeId = RedisKeyHelper.getBuilder().studentFileAnswerKey(req.getExamRecordDataId(), req.getOrder());
        ExamFileAnswer fileAnswer = examFileAnswerService.getFileAnswer(acknowledgeId);
        if (null == fileAnswer) {
            log.warn("{} is not exist.", acknowledgeId);
            throw new StatusException("100010", "无效的数据");
        }

        fileAnswer.setStatus(FileAnswerAcknowledgeStatus.valueOf(req.getAcknowledgeStatus()));
        examFileAnswerService.saveFileAnswer(acknowledgeId, fileAnswer);
    }

    @Deprecated
    @ApiOperation(value = "获取抓拍照片文件上传签名信息-又拍云（废弃）")
    @GetMapping("/getCapturePhotoUpYunSign")
    public FssSignInfo getCapturePhotoUpYunSign(GetYunSignDomainQuery query) {
        return getCaptureUploadSign(query);
    }

    @ApiOperation(value = "获取抓拍照片文件上传签名信息")
    @GetMapping("/getCapturePhotoYunSign")
    public FssSignInfo getCapturePhotoYunSign(GetYunSignDomainQuery query) {
        return getCaptureUploadSign(query);
    }

    private FssSignInfo getCaptureUploadSign(GetYunSignDomainQuery query) {
        User accessUser = this.getAccessUser();
        if (StringUtils.isNullOrEmpty(query.getFileSuffix())) {
            throw new StatusException("4001", "文件后缀名不允许为空");
        }

        String fileSuffix = query.getFileSuffix().startsWith(".") ? query.getFileSuffix() : "." + query.getFileSuffix();

        // 路径：/capture_photo/{rootOrgId}/{userId}/{timeMillis}{fileSuffix}
        String filePath = String.format(OeConstants.CAPTURE_PHOTO_PATH, accessUser.getRootOrgId(), accessUser.getUserId(), System.currentTimeMillis(), fileSuffix);
        FssSignInfo result = FssFactory.getInstance(false).buildSign(filePath);

        String signIdentifier = String.valueOf(System.currentTimeMillis());
        String signRedisKey = CacheConstants.CACHE_OE_UPYUN_IMAGE_SIGN + accessUser.getUserId() + "_" + signIdentifier;
        redisClient.set(signRedisKey, result, 60);
        result.setSignIdentifier(signIdentifier);

        return result;
    }

    @ApiOperation(value = "获取二维码")
    @PostMapping("/getQrCode")
    public String getQrCode(@RequestBody GetQrCodeReq req) {
        return examControlService.getQrCode(req, getAccessUser());
    }

    /**
     * 记录切换屏幕
     */
    @ApiOperation(value = "记录切换屏幕次数")
    @PostMapping("/switchScreen")
    public SwitchScreenCountInfo switchScreen(@RequestParam @ApiParam(value = "考试记录id") Long examRecordDataId) {
        User user = getAccessUser();
        String sequenceLockKey = CacheConstants.LOCK_EXAM_CONTROL + user.getUserId();
        // 开始考试上锁,分布式锁,系统在请求结束后会，自动释放锁，无需手动解锁
        SequenceLockHelper.getLock(sequenceLockKey);
        Check.isNull(examRecordDataId, "examRecordDataId不能为空");

        return examControlService.switchScreen(examRecordDataId);
    }

    @ApiOperation(value = "获取课程信息")
    @PostMapping("/getCourseInfo/{examRecordDataId}")
    public CourseInfo courseInfo(@PathVariable Long examRecordDataId) {
        return examRecordDataService.getCourseInfo(examRecordDataId);
    }

    @Deprecated
    @ApiOperation(value = "获取课程名称（废弃）")
    @GetMapping("/courseName/{examRecordDataId}")
    public String courseName(@PathVariable Long examRecordDataId) {
        CourseInfo info = examRecordDataService.getCourseInfo(examRecordDataId);
        if (info != null) {
            return info.getCourseName();
        }
        return null;
    }

}
